#include <u.h>
#include <libc.h>
#include <auth.h>
#include <bio.h>
#include <ndb.h>

typedef struct Nbdrequest Nbdrequest;
typedef struct Nbdreply Nbdreply;

enum
{
	Nbdrequestmagic = 0x25609513,
	Nbdreplymagic	= 0x67446698,
	Nbdread			= 0,
	Nbdwrite		= 1,
	Nbddisconnect	= 2,
};

struct Nbdrequest
{
	int		magic;
	int		type;
	uchar	handle[8];
	vlong	offset;
	int		len;
};

struct Nbdreply
{
	int		magic;
	int		error;
	uchar	handle[8];
};

int bufread(int fd, void *buf, int len)
{
	int c, l;
	l = 0;
	while(l < len){
		c = read(fd, &((uchar*)buf)[l], len-l);
		if(c < 0)
			return c;
		l += c;
	}
	return l;
}

int bufwrite(int fd, void *buf, int len)
{
	int c, l, u;
	l = 0;
	while(l < len){
		u = len - l;
		u = u > 1024 ? 1024 : u;
		c = write(fd, &((uchar*)buf)[l], u);
		if(c != u)
			return -1;
		l += c;
	}
	return l;
}

int readu32(int fd, int* v)
{
	uchar b[4];
	int ret;

	if((ret = bufread(fd, b, sizeof(b)))<0){
		return ret;
	}

	*v =
		((int)b[0]<<24)|
		((int)b[1]<<16)|
		((int)b[2]<< 8)|
		((int)b[3]<< 0);

	return 0;
}

int readu64(int fd, vlong* v)
{
	uchar b[8];
	int ret;

	if((ret = bufread(fd, b, sizeof(b)))<0){
		return ret;
	}

	*v =
		((vlong)b[0]<<56)|
		((vlong)b[1]<<48)|
		((vlong)b[2]<<40)|
		((vlong)b[3]<<32)|
		((vlong)b[4]<<24)|
		((vlong)b[5]<<16)|
		((vlong)b[6]<< 8)|
		((vlong)b[7]<< 0);

	return 0;
}

int writeu32(int fd, int v)
{
	uchar b[4];
	int ret;

	b[0] = (v & 0xff000000)>>24;
	b[1] = (v & 0x00ff0000)>>16;
	b[2] = (v & 0x0000ff00)>>8;
	b[3] = (v & 0x000000ff)>>0;

	if((ret = bufwrite(fd, b, sizeof(b)))<0){
		return ret;
	}

	return 0;
}

int writeu64(int fd, vlong v)
{
	uchar b[8];
	int ret;

	b[0] = (v&0xff00000000000000LL)>>56;
	b[1] = (v&0x00ff000000000000LL)>>48;
	b[2] = (v&0x0000ff0000000000LL)>>40;
	b[3] = (v&0x000000ff00000000LL)>>32;
	b[4] = (v&0x00000000ff000000LL)>>24;
	b[5] = (v&0x0000000000ff0000LL)>>16;
	b[6] = (v&0x000000000000ff00LL)>>8;
	b[7] = (v&0x00000000000000ffLL)>>0;

	if((ret = bufwrite(fd, b, sizeof(b)))<0){
		return ret;
	}

	return 0;
}

void
setuser(void)
{
	int fd;

	fd = open("#c/user", OWRITE);
	if(fd < 0 || write(fd, "none", strlen("none")) < 0)
		sysfatal("can't become none: %r");
	close(fd);
	if(newns("none", nil) < 0)
		sysfatal("can't build namespace: %r");
}

int nbdrequest(int fd, Nbdrequest* req)
{
	int ret;

	if((ret = readu32(fd, &req->magic)) < 0){
		return ret;
	}

	if((ret = readu32(fd, &req->type)) < 0){
		return ret;
	}

	if((ret = bufread(fd, &req->handle[0], sizeof(req->handle))) < 0){
		return ret;
	}

	if((ret = readu64(fd, &req->offset)) < 0){
		return ret;
	}

	if((ret = readu32(fd, &req->len)) < 0){
		return ret;
	}

	return 0;
}

int nbdreply(int fd, Nbdreply* rep)
{
	int ret;

	if((ret = writeu32(fd, rep->magic)) < 0)
		return ret;

	if((ret = writeu32(fd, rep->error)) < 0)
		return ret;

	if((ret = bufwrite(fd, rep->handle, sizeof(rep->handle))) < 0)
		return ret;

	return 0;
}

int nbdhello(int fd, vlong elen)
{
	int ret;
	uchar z[128];

	// INIT_PASSWD
	if((ret = bufwrite(fd, "NBDMAGIC", 8)) < 0){
		return ret;
	}

	// CLISERV_MAGIC
	if((ret = writeu64(fd, 0x00420281861253LL)) < 0){
		return ret;
	}

	// size
	if((ret = writeu64(fd, elen)) < 0){
		return ret;
	}

	// ZEROS
	memset(z, 0, 128);
	if((ret = bufwrite(fd, z, 128)) < 0){
		return ret;
	}
}

void nbdserve(int cfd, char *file)
{
	uchar		*buf;

	int		ffd;
	int		ret;
	Nbdrequest	req;
	Nbdreply	rep;
	Dir*		dir;
	uvlong		elen;

	buf = nil;

	ffd = open(file, ORDWR);
	if(ffd < 0){
		perror("open");
		goto outout;
	}
	dir = dirfstat(ffd);
	if(!dir){
		perror("dirstat");
		goto outout;
	}
	elen = dir->length;
	free(dir);

	if(nbdhello(cfd, elen) < 0){
		perror("nbdhello");
		goto outout;
	}

	while(1){
		ret = nbdrequest(cfd, &req);
		if(ret < 0){
			fprint(2, "closed\n");
			goto outout;
		}

		if(req.magic != Nbdrequestmagic){
			fprint(2, "invalid requst magic: 0x%x\n", req.magic);
			goto outout;
		}

		// repare reply
		rep.magic = Nbdreplymagic;
		memcpy(&rep.handle[0], &req.handle[0], 8);
		rep.error = 0;

		switch(req.type){
		case Nbdread:
			if((req.offset + req.len) > elen){
				fprint(2, "BUG: invalid len/offset\n");
				rep.error = -1;
				nbdreply(cfd, &rep);
				goto outout;
			}

			buf = malloc(req.len);
			if(buf==nil){
				fprint(2, "cant allocate buffer\n");
				rep.error = -1;
				nbdreply(cfd, &rep);
				goto outout;
			}

			seek(ffd, req.offset, 0);
			if(bufread(ffd, buf, req.len) < 0){
				perror("read file");
				rep.error = -1;
				nbdreply(cfd, &rep);
				goto outout;
			}

			rep.error = 0;
			nbdreply(cfd, &rep);

			if(bufwrite(cfd, buf, req.len) < 0){
				perror("write connection");
				goto outout;
			}
			free(buf);
			buf=nil;
			break;

		case Nbdwrite:
			if((req.offset + req.len) > elen){
				fprint(2, "BUG: invalid len/offset\n");
				rep.error = -1;
				nbdreply(cfd, &rep);
				goto outout;
			}

			buf = malloc(req.len);
			if(buf==nil){
				fprint(2, "cant allocate buffer\n");
				rep.error = -1;
				nbdreply(cfd, &rep);
				goto outout;
			}

			if(bufread(cfd, buf, req.len) < 0){
				perror("read client");
				rep.error = -1;
				nbdreply(cfd, &rep);
				goto outout;
			}

			seek(ffd, req.offset, 0);
			if(bufwrite(ffd, buf, req.len) < 0){
				perror("write file");
				rep.error = -1;
				nbdreply(cfd, &rep);
				goto outout;
			}

			rep.error = 0;
			nbdreply(cfd, &rep);

			free(buf);
			buf=nil;
			break;

		case Nbddisconnect:
			fprint(2, "disconnect\n");
			rep.error = 0;
			nbdreply(cfd, &rep);
			goto outout;

		default:
			fprint(2, "invalid request type: 0x%x\n", req.type);
			return;
		}
	}

outout:
	if(ffd >= 0)
		close(ffd);

	if(cfd >= 0)
		close(cfd);

	if(buf)
		free(buf);
}

char* nbdlookupfile(int lfd, char *ldir)
{
	char *ret;
	Ndb *db;
	Ndbtuple *t;
	NetConnInfo *ni;
	char *a[1] = { "nbdf" };

	t = nil;
	ret = nil;
	db = nil;

	ni = getnetconninfo(ldir, lfd);
	if(ni==nil)
		goto outout;

	db = ndbopen(nil);
	if(db==nil)
		goto outout;

	if(ni->rsys==nil)
		goto outout;

	t = ndbipinfo(db, "ip", ni->rsys, a, 1);
	if(t==nil)
		goto outout;

	if(t->val==nil)
		goto outout;

	ret = strdup(t->val);

outout:
	if(ni) freenetconninfo(ni);
	if(t) ndbfree(t);
	if(db) ndbclose(db);
	return ret;
}

void usage(void)
{
	fprint(2, "usage: nbdserver [-f file] [-a addr]\n");
	exits("usage");
}

void main(int argc, char **argv)
{
	int afd, lfd, dfd;
	char adir[256], ldir[256];
	char *f, *lf, *a;

	f=nil;
	a="tcp!*!4799";

	ARGBEGIN{
	case 'f':
		f = EARGF(usage());
		break;

	case 'a':
		a = EARGF(usage());
		break;

	default:
		usage();

	}ARGEND

	switch(rfork(RFNOTEG|RFPROC|RFFDG)){
	case -1:
		sysfatal("fork: %r");
	case 0:
		break;
	default:
		exits(0);
	}

	afd = announce(a, adir);
	if(afd < 0)
		sysfatal("announcing on %s: %r", adir);

	setuser();
	for(;;){
		lfd = listen(adir, ldir);
		if(lfd < 0){
			perror("listen");
			return;
		}

		switch(fork()){
		case -1:
			sysfatal("fork %r");

		case 0:
			lf = nbdlookupfile(lfd, ldir);
			if(lf == nil){
				lf = f?strdup(f):nil;
			}
			if(lf){
				dfd = accept(lfd, ldir);
				if(dfd >= 0)
					nbdserve(dfd, lf);
				free(lf);
			} else {
				reject(lfd, ldir, "no file for you");
			}
			exits(0);

		default:
			close(lfd);
			break;
		}
	}
}
